@@ -64,7 +64,7 @@ class DensityFlow(nn.Module):
                  z_dim: int = 50,
                  z_dist: Literal['normal','studentt','laplacian','cauchy','gumbel'] = 'studentt',
                  loss_func: Literal['negbinomial','poisson','multinomial','bernoulli'] = 'negbinomial',
-                 dispersion: float = 8.0,
+                 dispersion: float = 10.0,
                  use_zeroinflate: bool = False,
                  hidden_layers: list = [1024],
                  hidden_layer_activation: Literal['relu','softplus','leakyrelu','linear'] = 'relu',
@@ -270,16 +270,6 @@ class DensityFlow(nn.Module):
                     allow_broadcast=self.allow_broadcast,
                     use_cuda=self.use_cuda,
                 )
-        if self.loss_func == 'negbinomial':
-            self.decoder_dispersion = MLP(
-                    [self.latent_dim] + self.decoder_hidden_layers + [self.input_size],
-                    activation=activate_fct,
-                    output_activation=Exp,
-                    post_layer_fct=post_layer_fct,
-                    post_act_fct=post_act_fct,
-                    allow_broadcast=self.allow_broadcast,
-                    use_cuda=self.use_cuda,
-                )
 
         if self.latent_dist == 'studentt':
             self.codebook = MLP(
@@ -444,9 +434,9 @@ class DensityFlow(nn.Module):
         batch_size = xs.size(0)
         self.options = dict(dtype=xs.dtype, device=xs.device)
         
-        #if self.loss_func=='negbinomial':
-        #    dispersion = pyro.param("dispersion", self.dispersion *
-        #                                    xs.new_ones(self.input_size), constraint=constraints.positive)
+        if self.loss_func=='negbinomial':
+            dispersion = pyro.param("dispersion", self.dispersion *
+                                            xs.new_ones(self.input_size), constraint=constraints.positive)
             
         if self.use_zeroinflate:
             gate_logits = pyro.param("dropout_rate", xs.new_zeros(self.input_size))
@@ -490,7 +480,6 @@ class DensityFlow(nn.Module):
                 log_theta = log_mu
             elif self.loss_func == 'negbinomial':
                 mu = log_mu.exp()
-                dispersion = self.decoder_dispersion(zs)
             else:
                 rate = log_mu.exp()
                 theta = dist.DirichletMultinomial(total_count=1, concentration=rate).mean
@@ -498,8 +487,8 @@ class DensityFlow(nn.Module):
                     rate = theta * torch.sum(xs, dim=1, keepdim=True)
 
             if self.loss_func == 'negbinomial':
-                #logits = (mu.log()-dispersion.log()).clamp(min=-15, max=15)
-                logits = (dispersion.log()-mu.log()).clamp(min=-15, max=15)
+                logits = (mu.log()-dispersion.log()).clamp(min=-15, max=15)
+                #logits = (dispersion.log()-mu.log()).clamp(min=-15, max=15)
                 if self.use_zeroinflate:
                     pyro.sample('x', dist.ZeroInflatedDistribution(dist.NegativeBinomial(total_count=dispersion, logits=logits),gate_logits=gate_logits).to_event(1), obs=xs)
                 else:
@@ -907,25 +896,22 @@ class DensityFlow(nn.Module):
         A = np.concatenate(A)
         return A
     
-    def predict(self, xs, us, perturbs_predict:list, perturbs_reference:list, library_sizes=None, show_progress=True):
+    def predict(self, zs, perturbs_predict:list, perturbs_reference:list, library_sizes=None, show_progress=True):
         perturbs_reference = np.array(perturbs_reference)
+        perturbs_predict = np.unique(perturbs_predict)
         
         # basal embedding
-        zs = self.get_basal_embedding(xs, show_progress=show_progress)
+        dzs_total = np.zeros_like(zs)
         for pert in perturbs_predict:
-            pert_idx = int(np.where(perturbs_reference==pert)[0])
-            us_i = us[:,pert_idx].reshape(-1,1)
+            if pert in perturbs_reference:
+                pert_idx = int(np.where(perturbs_reference==pert)[0])
             
-            # factor effect of xs 
-            dzs0 = self.get_cell_shift(zs, perturb_idx=pert_idx, perturb_us=us_i, show_progress=show_progress)
-            
-            # perturbation effect 
-            ps = np.ones_like(us_i)
-            if np.sum(np.abs(ps-us_i))>=1:
+                # perturbation effect 
+                ps = np.ones([zs.shape[0],1])
                 dzs = self.get_cell_shift(zs, perturb_idx=pert_idx, perturb_us=ps, show_progress=show_progress)          
-                zs = zs + dzs0 + dzs
-            else:
-                zs = zs + dzs0
+                dzs_total = dzs_total + dzs
+        
+        zs = zs + dzs_total
         
         if library_sizes is None:
             library_sizes = np.sum(xs, axis=1, keepdims=True)
@@ -1008,18 +994,15 @@ class DensityFlow(nn.Module):
     def _count(self, log_mu, library_size=None):
         if self.loss_func == 'bernoulli':
             counts = dist.Bernoulli(logits=log_mu).to_event(1).mean
-        elif self.loss_func == 'multinomial':
-            theta = dist.Multinomial(total_count=int(1e8), logits=log_mu).mean
-            counts = theta * library_size
+        elif self.loss_func == 'negbinomial':
+            counts = log_mu.exp()
         else:
             rate = log_mu.exp()
             theta = dist.DirichletMultinomial(total_count=1, concentration=rate).mean
             counts = theta * library_size
         return counts
     
-    def get_counts(self, zs, library_sizes,
-                        batch_size: int = 1024,
-                        show_progress=True):
+    def get_counts(self, zs, library_sizes, batch_size: int = 1024, show_progress=True):
 
         zs = convert_to_tensor(zs, device=self.get_device())
         
@@ -1028,7 +1011,7 @@ class DensityFlow(nn.Module):
         elif len(library_sizes.shape)==1:
             library_sizes = library_sizes.reshape(-1,1)
         ls = convert_to_tensor(library_sizes, device=self.get_device())
-            
+        
         dataset = CustomDataset2(zs,ls)
         dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False)
         
@@ -1048,7 +1031,7 @@ class DensityFlow(nn.Module):
             ad = sc.AnnData(xs)
             binarize(ad, threshold=threshold)
             xs = ad.X.copy()
-        else:
+        elif self.loss_func in ['poisson','multinomial']:
             xs = np.round(xs)
             
         if sparse.issparse(xs):
